"""This module contains the LogFile class, an encapsulation of log variables."""

import re
from heybrochecklog.resources import DEDUCTIONS


class LogFile:
    """A log file class containing variables, score, deductions, etc."""

    def __init__(self, contents, ripper=None):

        self.full_contents = contents
        self.contents = format_full_contents(contents)
        self.concat_contents = [line for line in self.contents if line.strip()]
        self.score = 100
        self.ripper = ripper
        self.language = None
        self.drive = None
        self.version = None
        self.album = None
        self.unrecognized = None

        # Some other log settings
        self.range = False
        self.cdr = False
        self.unindexed_drive = False
        self.htoa = False
        self.htoa_index = False
        self.htoa_ripped = False

        # Important parts of the log
        self.checksum = False
        self.all_tracks = None
        self.deductions = {}
        self.crc_mismatch = []
        self.track_errors = {
            "Aborted copy": [],
            "Timing problem": [],
            "Suspicious position": [],
            "Missing samples": [],
            "Read error": [],
            "Damaged sector count": []
        }

        # Lists of data for the log
        self.toc = {}
        self.accuraterip = []
        self.track_indices = []
        self.tracks = {}

        # Indexes of log locations
        self.index_settings = None
        self.index_toc = None
        self.index_tracks = None
        self.index_footer = None

        # Flagged = auto report log
        self.flagged = False

    def to_dict(self):
        """Return a dict of the log analysis."""
        if self.unrecognized:
            return {
                'unrecognized': self.unrecognized,
                'flagged': self.flagged,
                'contents': ''.join(self.full_contents)
            }

        deductions = [deduction for deduction in self.deductions.values()]

        return {
            'deductions': deductions,
            'flagged': self.flagged,
            'name': self.album,
            'ripper': self.ripper,
            'score': self.score,
            'version': self.version,
            'unrecognized': False,
            'contents': ''.join(self.full_contents)
        }

    def add_deduction(self, deduction, multiplier=1, track=None, extra_phrase=None):
        """Add a deduction to the log file."""
        name, score = self._get_deduction_from_dict(deduction)
        if score:
            score = score * min(10, multiplier)

        if track:
            name = 'Track {}: {}'.format(track, name)
        if multiplier > 1:
            name += ' ({} occurrences)'.format(multiplier)
        if score:
            name += ' (-{} points)'.format(score)
        if extra_phrase:
            name += ' ({})'.format(extra_phrase)

        self.deductions[deduction] = [name, score]

    def _get_deduction_from_dict(self, deduction):
        """Get the deduction's name and score from the deductions dict."""
        if deduction not in DEDUCTIONS:
            return (deduction, None)

        deduction_entry = DEDUCTIONS[deduction]
        # Some deductions are different per-ripper and are represented
        # with an extra embedded dictionary.
        if isinstance(deduction_entry, dict):
            if self.ripper in deduction_entry:
                deduction_entry = deduction_entry[self.ripper]
            else:
                deduction_entry = deduction_entry['Default']

        return (deduction_entry[0], deduction_entry[1])

    def remove_deduction(self, deduction):
        """Remove a deduction from the log file."""
        if deduction in self.deductions:
            del self.deductions[deduction]

    def has_deduction(self, deduction):
        """Learn whether or not the log file has a deduction."""
        return deduction in self.deductions

    def has_deductions(self, *deductions):
        """Return whether or not log has every deduction in deductions."""
        return all(de in self.deductions for de in deductions)


def format_full_contents(full_contents):
    """Format raw contents by stripping spaces, blank lines, and filtering out unicode crap."""
    contents = [re.sub(r'\s+', ' ', l.rstrip()) for l in full_contents]
    contents = [re.sub('：', ':', l) for l in contents]
    contents = [re.sub('，', ', ', l) for l in contents]

    return contents



"""This module contains shared functions between the various top-level modules."""

import os
import json
import codecs
import chardet


def get_log_contents(log_file):
    """Open a log file and return its contents."""
    encoding = get_log_encoding(log_file)
    with log_file.open(encoding=encoding) as log:
        contents = log.readlines()

    return contents


def get_log_encoding(log_file):
    """Get the encoding of the log file with the chardet library."""
    raw = log_file.read_bytes()
    if raw.startswith(codecs.BOM_UTF8):
        return 'utf-8-sig'
    else:
        result = chardet.detect(raw)
        return result['encoding'] if result['confidence'] > 0.7 else 'utf-8-sig'


def format_pattern(pattern, append=None):
    if append:
        pattern = [p + append for p in pattern]
    return '|'.join(pattern)


def open_json(*paths):
    """Open the language JSON patterns file and return it."""
    basepath = get_path()
    with open(os.path.join(basepath, 'resources', *paths)) as jsonfile:
        language_data = json.load(jsonfile)

    return language_data


def get_path():
    """Get the filepath for the heybrochecklog package directory."""
    return os.path.abspath(os.path.dirname(__file__))



"""This module handles the log scoring functionality of the heybrochecklog package."""

from heybrochecklog import UnrecognizedException
from heybrochecklog.analyze import analyze_log
from heybrochecklog.shared import get_log_contents, open_json
from heybrochecklog.logfile import LogFile
from heybrochecklog.score import eac, xld, eac95


def score_log(log_file, markup=False):
    try:
        contents = get_log_contents(log_file)
        log = LogFile(contents)
        log = score_wrapper(log, markup)
    except UnicodeDecodeError:
        log = LogFile('')
        log.unrecognized = 'Could not decode log file.'
    return log.to_dict()


def score_log_from_contents(contents):
    """Score a log file given its contents, instead of opening it from a file."""
    log = LogFile(contents.split('\n'))
    try:
        log = score_wrapper(log)
    except UnicodeDecodeError:
        log.unrecognized = 'Could not decode log file.'
    return log.to_dict()


def score_wrapper(log, markup=False):
    """Determine the type of log file and passes the log to the appropriate logchecker."""

    try:
        analyze_log(log)
    except UnrecognizedException as exception:
        log.unrecognized = str(exception)
        return log

    if log.ripper == 'EAC':
        info_json = open_json('eac', '{}.json'.format(log.language))
        logchecker = eac.EACChecker(info_json['patterns'], info_json['translation'], markup)
    elif log.ripper == 'XLD':
        patterns = open_json('xld.json')
        logchecker = xld.XLDChecker(patterns, markup=markup)
    elif log.ripper == 'EAC95':
        info_json = open_json('eac95', '{}.json'.format(log.language))
        logchecker = eac95.EAC95Checker(info_json['patterns'], info_json['translation'], markup)

    try:
        log = logchecker.check(log)
    except UnrecognizedException as exception:
        log.unrecognized = str(exception)

    return log



"""Mark up the log file with classes according to the regex."""

import re
import html
from heybrochecklog.shared import format_pattern as fmt_ptn
from heybrochecklog.score.modules import parsers
from heybrochecklog.markup.matches import (eac_track_matches, xld_track_matches, eac_footer_matches,
                                           xld_footer_matches, xld_ar_summary)

VERSIONS = {
    'EAC': r'Exact Audio Copy (V.*) from (.*)',
    'XLD': r'X Lossless Decoder version ([0-9abc]+) \(([0-9\.]+)\)'
}


def markup(log, patterns, translation):
    """Mark up log files with highlighting for proper/improper settings."""
    log.full_contents = [html.escape(line) for line in log.full_contents]

    header(log, translation)
    settings(log, patterns)
    toc(log)
    tracks(log, patterns, translation)
    footer(log, translation)


def header(log, translation):
    """Mark up the header of the log."""
    # If first line is a version line, style it.
    start_index = 0
    if log.ripper in VERSIONS:
        log.full_contents[0] = substitute(log.full_contents[0], VERSIONS[log.ripper], 'log1')
        start_index = 1  # No need to scan this line twice.

    if log.ripper == 'EAC' or log.ripper == 'EAC95':
        re_time_line = translation['1274']
    else:  # XLD
        re_time_line = 'XLD extraction logfile from'

    count = 1
    for i, line in enumerate(log.full_contents[start_index:log.index_settings]):
        i += start_index
        # Style the extraction time line.
        if line.strip():
            if count == 1:
                line = substitute(line, '{} (.*)'.format(re_time_line), 'log5')
                log.full_contents[i] = substitute(line, '({}.*)'.format(re_time_line), 'good')
                count += 1
            # Style the Artist / Album string
            elif count == 2:
                log.full_contents[i] = substitute(line, '(.*)', 'log4')
                count += 1
            # Style drive line
            elif count == 3:
                log.full_contents[i] = drive(log, line)
                count += 1
            # XLD CD Type line
            elif count == 4:
                log.full_contents[i] = cd_type(log, line)
                break


def drive(log, line):
    """Mark up the drive line according to offset."""
    if log.has_deduction('Virtual drive'):
        return style_setting(line, 'bad')
    elif log.unindexed_drive:
        line = line.rstrip()
        line += ' (not found in database)\n'
        return style_setting(line, 'badish')
    return style_setting(line, 'good')


def cd_type(log, line):
    """Style CD type line."""
    if log.cdr:
        return style_setting(line, 'badish')
    return style_setting(line, 'good')


def settings(log, patterns):
    """Mark up the settings block."""
    for i, line in enumerate(log.full_contents[log.index_settings:log.index_toc]):
        i += log.index_settings
        if re.match(fmt_ptn(patterns['settings']['Read mode']), line) and log.ripper == 'EAC95':
            log.full_contents[i] = style_95_read_mode(line, patterns)
            continue
        for key, value in patterns['settings'].items():
            if re.match(fmt_ptn(value), line.lstrip()):  # lstrip() for EAC95
                class_ = 'bad' if log.has_deduction(key) else 'good'
                log.full_contents[i] = style_setting(line, class_)
                break
        else:
            if log.ripper in ['EAC', 'EAC95']:
                for key, value in patterns['bad settings'].items():
                    if re.match(fmt_ptn(value), line.lstrip()):
                        log.full_contents[i] = style_setting(line, 'bad')
                        break
                else:
                    log.full_contents[i] = style_setting(line, 'log4')
            else:
                log.full_contents[i] = style_setting(line, 'log4')


def toc(log):
    """Mark up TOC block."""
    # Mark up beginning of block
    init_line = log.full_contents[log.index_toc]
    log.full_contents[log.index_toc] = substitute(init_line, '(.*)', 'log4 log5')

    # Adjust line for XLD All Tracks block
    end_line = log.all_tracks if log.all_tracks else log.index_tracks
    if log.ripper == 'XLD':  # XLD AR Summary block matches
        matches = xld_ar_summary()

    re_toc_title = re.compile(r' +[^0-9]+ +\| +[^0-9]+ +\| +[^0-9]+ +\| +[^0-9]+ +\| +[^0-9]+ *$')
    re_toc_entry = r' +[0-9]+ +\| +([0-9:\.]+) +\| +([0-9:\.]+) +\| +([0-9]+) +\| +([0-9]+) *$'
    for i, line in enumerate(log.full_contents[log.index_toc:end_line]):
        i += log.index_toc
        if re_toc_title.match(line):
            log.full_contents[i] = sub_strong(line, '(.*)')
        elif re.match(r'\s+-+\s*$', line):
            log.full_contents[i] = sub_strong(line, '(-+)')
        elif re.match(re_toc_entry, line):
            log.full_contents[i] = sub_toc(line)

        # XLD also has an AR Summary block.
        if log.ripper == 'XLD':
            for class_ in matches:
                for element in matches[class_]:
                    if re.match(element, line.lstrip()):
                        log.full_contents[i] = substitute(line, '({})'.format(element), class_)


def tracks(log, patterns, translation):
    """Mark up the tracks block."""
    if log.ripper == 'XLD':
        xld_tracks(log, patterns)
    else:
        eac_tracks(log, patterns, translation)


def xld_tracks(log, patterns):
    """XLD tracks."""
    matches = xld_track_matches()
    indices = [log.all_tracks] + log.track_indices if log.all_tracks else log.track_indices

    for i, index in enumerate(indices):
        track_num, log.full_contents[index] = track_number(log, index, patterns['track'])
        for j, line in enumerate(log.full_contents[indices[i]:indices[i + 1]]):
            j += indices[i]
            for match in matches['full_line']:
                if re.match(match[1], line.lstrip()):
                    log.full_contents[j] = substitute(line, '({}.*)'.format(match[1]), match[0])
                    break
            else:
                for match in matches['crc']:
                    if re.match(match + ' +:', line.lstrip()) and track_num:
                        track = log.tracks[track_num]
                        log.full_contents[j] = sub_crc(track, match, line, xld_colon=True)
                        break
                else:
                    found = False
                    for class_ in matches['statistics']:
                        for element in matches['statistics'][class_]:
                            if re.match(element, line.lstrip()):
                                log.full_contents[j] = style_statistic(line, class_)
                                found = True
                                break
                        if found:
                            break
                    else:
                        log.full_contents[j] = style_setting(line, 'log3', first_class='log4',
                                                             include_colon=True)

        if indices[i + 1] == max(indices):
            break


def eac_tracks(log, patterns, translation):
    """EAC tracks."""
    matches = eac_track_matches(translation)

    for i, index in enumerate(log.track_indices):
        track_num, log.full_contents[index] = track_number(log, index, patterns['track'])
        for j, line in enumerate(log.full_contents[log.track_indices[i]:log.track_indices[i + 1]]):
            j += log.track_indices[i]
            for element in matches['log4']:
                if re.match(element, line.lstrip()):
                    line = substitute(line, ' +{} +(.*)'.format(element), 'log3')
                    log.full_contents[j] = substitute(line, ' +({}.+)'.format(element), 'log4')
                    break
            for class_ in ['good', 'badish', 'bad', 'log3']:
                for element in matches[class_]:
                    if re.match(element, line.lstrip()):
                        log.full_contents[j] = substitute(line, ' *({}.*)'.format(element), class_)
                        break
            for element in matches['crc']:
                if re.match(element, line.lstrip()):
                    track = log.tracks[track_num]
                    log.full_contents[j] = sub_crc(track, element, line)
        if log.track_indices[i + 1] == max(log.track_indices):
            break


def track_number(log, index, track_pattern):
    """Parse track number and style the line."""
    if log.all_tracks and log.full_contents[index].startswith('All Tracks'):
        return (0, substitute(log.full_contents[index], '(.*)', 'log5'))
    track_num = parsers.get_track_number(log, index, track_pattern)
    line = substitute(log.full_contents[index], r'{} +(\d+)'.format(track_pattern), 'log4 log1')
    line = substitute(line, '({}.+)'.format(track_pattern), 'log5')
    return (track_num, line)


def sub_crc(track, element, line, xld_colon=False):
    """Process the CRCs for markup."""
    re_crc = '([0-9A-F]{8})'
    if 'test crc' not in track:
        line = substitute(line, re_crc, 'badish')
    elif track['test crc'] != track['copy crc']:
        line = substitute(line, re_crc, 'bad')
    else:
        line = substitute(line, re_crc, 'good')
    if xld_colon:
        return substitute(line, ' +({}.+ :)'.format(element), 'log4')
    return substitute(line, ' +({}.+)'.format(element), 'log4',)


def footer(log, translation):
    """Mark up the footer."""
    matches = xld_footer_matches() if log.ripper == 'XLD' else eac_footer_matches(translation)

    for i, line in enumerate(log.full_contents[log.index_footer:]):
        i += log.index_footer
        for class_ in matches:
            for element in matches[class_]:
                if re.match(element, line.lstrip()):
                    log.full_contents[i] = substitute(line, '({})'.format(element), class_)
                    break
            if log.ripper == 'XLD':
                # XLD Checksum stuff goes here
                if line.startswith('-----BEGIN XLD SIGNATURE-----'):
                    log.full_contents[i] = '<span class="good">-----BEGIN XLD SIGNATURE-----\n'
                elif line.startswith('-----END XLD SIGNATURE-----'):
                    log.full_contents[i] = '-----END XLD SIGNATURE-----</span>'


def style_setting(line, class_, first_class='log5', include_colon=False):
    """Style a setting line in the log (<setting_name> +: <setting>)."""
    if re.match(r'.+:.+', line):
        parts = line.split(':', 1)
        setting = re.escape(parts[0].lstrip() + ':' if include_colon else parts[0].lstrip())
        data = '({})'.format(re.escape(parts[1].strip()))
        line = substitute(line, '({})'.format(setting), first_class)
        line = substitute(line, data, class_)
    return line


def style_statistic(line, class_):
    """Style a XLD statistic line."""
    if re.match(r'.+:.+', line):
        parts = line.split(':', 1)
        statistic = re.escape(parts[0].lstrip() + ':')
        occurrences = parts[1].strip()

        if occurrences.isdigit() and int(occurrences) == 0:
            class_ = 'good'

        line = substitute(line, '({})'.format(occurrences), class_)
        line = substitute(line, '({})'.format(statistic), 'log4')

    return line


def style_95_read_mode(line, patterns):
    """Style the EAC 95 read mode line."""
    # Burst mode doesn't have multiple settings in one line
    if ',' not in line:
        return style_setting(line, 'bad')

    split_line = line.split(':', 1)

    read_mode = split_line[0].rstrip()
    line = line.replace(read_mode, '<span class="log5">{}</span>'.format(read_mode), 1)

    parts = split_line[1].lstrip().split(' ', 1)
    parts[1:] = [part.strip() for part in parts[1].split(',')]
    num = 0
    p = patterns['95 settings']
    for setting in [p['Read mode'], p['C2 pointers'], p['Accurate stream'], p['Audio cache']]:
        if num == len(parts):
            break
        class_ = 'good' if setting in line else 'bad'
        line = line.replace(parts[num], '<span class="{}">{}</span>'.format(class_, parts[num]), 1)
        num += 1

    return line


def substitute(line, regex, class_):
    """Generate a span and put it in the line per the regex."""
    result = re.search(regex, line)
    if result:
        for match in result.groups():
            line = line.replace(match, '<span class="{}">{}</span>'.format(class_, match), 1)
    return line


def sub_strong(line, regex):
    """Surround some patterns in <strong> tags."""
    result = re.search(regex, line)
    if result:
        for match in result.groups():
            line = line.replace(match, '<strong>{}</strong>'.format(match), 1)
    return line


def sub_toc(line):
    """Substitute the HTML markup into a line of the TOC."""
    return re.sub(r'([0-9]+)( +)\|( +)([0-9:\.]+)( +)\|( +)([0-9:\.]+)( +)\|( +)'
                  r'([0-9]+)( +)\|( +)([0-9]+)',
                  r'<span class="log4">\1</span>\2<strong>|</strong>'
                  r'\3<span class="log1">\4</span>\5<strong>|</strong>'
                  r'\6<span class="log1">\7</span>\8<strong>|</strong>'
                  r'\9<span class="log1">\10</span>\11<strong>|</strong>'
                  r'\12<span class="log1">\13</span>', line)


def re_paren(line):
    """Regex the comma. Quality docstring."""
    return re.sub(r'\(', r'\(', line)



"""This module analyzes the log file and determines generic stuff like it's ripper and language, etc.
"""

import re
from heybrochecklog import UnrecognizedException
from heybrochecklog.resources import EAC_RIPLINES


def analyze_log(log):
    """Analyze a log file and determine some generic background information."""
    log.ripper = get_ripper(log.contents)
    log.language = determine_language(log)


def get_ripper(contents):
    """Determine the ripper used in the log."""
    if not contents:  # Is file empty?
        raise UnrecognizedException('Empty log file')

    # Even foreign EAC logs start with this line.
    eac_regex = re.compile(r'Exact Audio Copy V[0-1]\.[0-9]+.*?from.*')
    xld_regex = re.compile(r'X Lossless Decoder version ([0-9abc]+) \([0-9\.]+\)')
    if eac_regex.match(contents[0]):
        return 'EAC'
    elif xld_regex.match(contents[0]):
        return 'XLD'
    else:
        # Unfortunately, not all EAC <=0.95 logs are English, so a compiled multi-language ripper
        # regex pattern is necessary.
        re_95 = re.compile('|'.join(list(EAC_RIPLINES.values())))
        if re_95.match(contents[0]):
            return 'EAC95'
        else:
            raise UnrecognizedException('Unrecognized ripper')


def determine_language(log):
    """Determine the language of the log file, and verify that it is an EAC log file."""
    if log.ripper == 'XLD':
        return 'english'

    useful_contents = [re.sub(r'\s+', ' ', l.rstrip()) for l in log.contents if l.strip()]
    for line in useful_contents[:2]:
        for language, line_starter in EAC_RIPLINES.items():
            if re.match(line_starter, line):
                return language

    raise UnrecognizedException('Unrecognized/unsupported language')



"""This module contains the XLD Log Checker."""

import re
from heybrochecklog.markup import markup
from heybrochecklog import UnrecognizedException
from heybrochecklog.shared import format_pattern as fmt_ptn
from heybrochecklog.score.logchecker import LogChecker
from heybrochecklog.score.modules import parsers, validation


class XLDChecker(LogChecker):
    """This class analyzes XLD Log Files."""

    def check(self, log):
        """Checks the XLD logs."""
        if len(log.contents) < 25:
            raise UnrecognizedException('Cannot parse log file; log file too short')

        log.version = self.check_version(log)
        log.album = log.concat_contents[2]
        log.drive = self.check_drive(log)
        self.check_cdr(log)

        self.index_log(log)
        self.evaluate_settings(log)
        parsers.index_toc(log)
        self.check_tracks(log)
        self.is_there_a_htoa(log)
        validation.validate_track_count(log)
        validation.validate_track_settings(log, xld=True)
        parsers.parse_checksum(log, self.patterns['checksum'], '20121222', 'XLD pre-142.2')

        self.deduct_and_score(log)
        if self.markup:
            markup(log, self.patterns, self.translation)

        return log

    def check_version(self, log):
        """Check the version of the log and verify it is acceptable."""
        regex = re.compile(r'X Lossless Decoder version ([0-9abc]+) \(([0-9\.]+)\)')
        return self.verify_version(regex, log.concat_contents[0], 'XLD')

    def check_drive(self, log):
        """Check the drive of the log and verify it is an allowed drive."""
        regex = r' *: (.*)?(?: +\(revision [A-Z0-9\.]\))?$'
        return self.get_drive(regex, log.concat_contents[3])

    def check_cdr(self, log):
        """Check the log to see if CD-R is flagged."""
        result = re.search(fmt_ptn(self.patterns['disc type']) + r' : (.*)', log.concat_contents[4])
        if result:
            if result.group(1) == 'Pressed CD':
                return
            elif result.group(1) == 'CD-Recordable':
                log.cdr = True
                log.flagged = True
                log.add_deduction('CD-R')
            else:
                raise UnrecognizedException('Unknown disc type')

    def all_range_index(self, log, line):
        """Match the Range Rip line in the log file."""
        if log.all_tracks is None and re.match(fmt_ptn(self.patterns['All Tracks']), line):
            return True
        return False

    def all_range_index_action(self, log, line_num):
        """Action to take when the range rip line is matched."""
        log.all_tracks = line_num

    def is_there_a_htoa(self, log):
        """Check rip for Hidden Track One Audio."""
        # 450 sectors or 6 seconds minimum, one track rip with pregap (containing HTOA)
        # appended to the first track. It is then split from the first track with
        # Audacity/Audition.
        if len(log.tracks) == 1 and log.toc[1][0] >= 450:
            for line in log.contents[log.index_settings:log.index_toc]:
                if re.match(r'Gap status +: Analyzed, Appended$', line):
                    log.add_deduction('HTOA extracted')
                    break

    def check_tracks(self, log):
        """Get track data for each track and check for errors."""
        tsettings = self.patterns['track settings']
        track_settings = {
            'filename': re.compile(r'\s+' + fmt_ptn(tsettings['filename']) + r' : (.*?\/.*?\..*)'),
            'pregap': re.compile(r'\s+' + fmt_ptn(tsettings['pregap']) + r' : ([0-9:\.]+)'),
            'gain': re.compile(r'\s+' + fmt_ptn(tsettings['gain']) + r' : ([A-Za-z0-9\.-]+)'),
            'peak': re.compile(r'\s+' + fmt_ptn(tsettings['peak']) + r' : ([0-9\.]+)'),
            'test crc': re.compile(r'\s+' + fmt_ptn(tsettings['test crc']) + r' : ([A-Z0-9]{8})'),
            'copy crc': re.compile(r'\s+' + fmt_ptn(tsettings['copy crc']) + r' : ([A-Z0-9]{8})')
        }

        if log.all_tracks:
            for i in range(log.all_tracks, min(log.track_indices)):
                if re.match(track_settings['filename'], log.contents[i]):
                    log.range = True
                    break

        self.analyze_tracks(log, track_settings, parsers.parse_errors_xld)

    def evaluate_tracks(self, log):
        """Evaluate the analyzed track data for deficiencies (actually split off the
        logchecker-specific) stuff ;)
        """
        # Deduct for a Range Rip
        if log.range:
            log.add_deduction('Range rip')

        # Check AccurateRip - Mismatching AR results can indicate problems even with T&C
        validation.analyze_accuraterip(log)

    def deduct_and_score(self, log):
        """Process the accumulated deductions and score the log file."""
        # Check for presence of Track gain in the tracks.
        if not all('gain' in log.tracks[track] for track in log.tracks):
            log.add_deduction('Track gain')

        # Deduct for accumulated per-track XLD deductions; since XLD accumulated deductions are
        # deducted by an occurrence basis, we are splitting deductions into one deduction per track,
        # capped at 10%.
        for error in log.track_errors:
            for each in log.track_errors[error]:
                log.add_deduction(error, multiplier=each[1], track=each[0])

        super().deduct_and_score(log)



"""This module contains the EAC Log Checker."""

import re
from heybrochecklog.markup import markup
from heybrochecklog import UnrecognizedException
from heybrochecklog.shared import format_pattern as fmt_ptn
from heybrochecklog.score.logchecker import LogChecker
from heybrochecklog.score.modules import combined, parsers, validation


class EACChecker(LogChecker):
    """This class analyzes >0.95 EAC Log Files."""

    def check(self, main_log):
        """Checks the EAC logs."""
        logs = combined.split_combined(main_log)
        for log in logs:
            if len(log.concat_contents) < 20:
                raise UnrecognizedException('Cannot parse log file; log file too short')

            log.version = self.check_version(log)
            log.album = log.concat_contents[2]
            log.drive = self.check_drive(log)

            self.index_log(log)
            self.evaluate_settings(log)
            parsers.index_toc(log)
            self.is_there_a_htoa(log)
            self.check_tracks(log)
            parsers.parse_checksum(log, self.patterns['checksum'], 'V1.0 beta 1', 'EAC <1.0')
            if self.markup:
                markup(log, self.patterns, self.translation)

        main_log = combined.defragment(logs)
        validation.validate_track_count(main_log)
        validation.validate_track_settings(main_log)
        self.deduct_and_score(main_log)

        return main_log

    def check_version(self, log):
        """Check the version of the log and verify it is acceptable."""
        regex = re.compile(r'Exact Audio Copy (V.*) from (.*)')
        return self.verify_version(regex, log.concat_contents[0], 'EAC')

    def check_drive(self, log):
        """Check the drive of the log and verify it is an allowed drive."""
        regex = r' ?: (.*) Adapter:[ 0-9]+ID:[ 0-9]+$'
        return self.get_drive(regex, log.concat_contents[3])

    def all_range_index(self, log, line):
        """Match the Range Rip line in the log file."""
        if log.index_tracks is None and re.match(fmt_ptn(self.patterns['range']), line):
            return True
        return False

    def all_range_index_action(self, log, line_num):
        """Action to take when the range rip line is matched."""
        log.track_indices.append(line_num)
        log.range = True

    def check_bad_settings(self, log, line):
        """Evaluate the instant -100 point deductions."""
        bad_settings = self.patterns['bad settings']
        for sett, pattern in bad_settings.items():
            if re.match(fmt_ptn(pattern), line):
                log.add_deduction(sett)

    def evaluate_unmatched_settings(self, log, settings):
        """Override super to account for burst mode not having some settings."""
        burst_no_exist = ['Accurate stream', 'Audio cache', 'C2 pointers']
        if log.has_deduction('Read mode'):
            if any(setting not in settings for setting in burst_no_exist):
                raise UnrecognizedException('Invalid rip settings for a burst/fast mode rip')
            for setting in burst_no_exist:
                del settings[setting]

        if log.has_deduction('Combined offset') and 'Drive offset' in settings:
            del settings['Drive offset']

        super().evaluate_unmatched_settings(log, settings)

    def is_there_a_htoa(self, log):
        """Check rip for Hidden Track One Audio."""
        # 6 second minimum for HTOA per EAC standards
        # Only accepted HTOA extraction technique for EAC is range-based
        if log.toc[1][0] < 450 or not log.range:
            return

        for line in log.contents[log.index_tracks + 1:]:
            if line.strip():
                result = re.search(fmt_ptn(self.patterns['htoa']), line)
                if result:
                    log.htoa = True
                    log.htoa_index = log.toc[1][0] - 1

                    # Remove the gap handling notification (doesn't appear for range rips)
                    if log.has_deduction('Could not verify gap handling'):
                        log.remove_deduction('Could not verify gap handling')

                    if int(result.group(1)) == log.htoa_index:
                        log.htoa_ripped = True
                        log.add_deduction('HTOA extracted')
                    else:
                        log.add_deduction('Improper HTOA extraction')
                else:
                    log.add_deduction('HTOA detected, but not extracted')
                break

    def check_tracks(self, log):
        """Wrapper for the analyze_tracks method. Get track data for every track
        and check for errors.
        """
        tsettings = self.patterns['track settings']
        track_settings = {
            'filename': re.compile(r'\s+' + fmt_ptn(tsettings['filename']) + r' (.*)'),
            'pregap': re.compile(r'\s+' + fmt_ptn(tsettings['pregap']) + r' ([0-9:\.]+)'),
            'peak': re.compile(r'\s+' + fmt_ptn(tsettings['peak']) + r' ([0-9\.]+) %'),
            'test crc': re.compile(r'\s+' + fmt_ptn(tsettings['test crc']) + r' ([A-Z0-9]{8})'),
            'copy crc': re.compile(r'\s+' + fmt_ptn(tsettings['copy crc']) + r' ([A-Z0-9]{8})')
        }

        self.analyze_tracks(log, track_settings, parsers.parse_errors_eac)

    def evaluate_tracks(self, log):
        """Evaluate the analyzed track data for deficiencies."""
        # AccurateRip for EAC Range Rip - AR results are at the bottom of the log.
        if log.range:
            patterns = self.patterns['range accuraterip'].items()
            parsers.parse_range_accuraterip(log, patterns)

        # HTOA doesn't need these deductions, since it's supposed to be range ripping.
        if not log.htoa:
            if log.range:
                log.add_deduction('Range rip')
            # Check AccurateRip - Mismatching AR results can indicate problems even with T&C
            validation.analyze_accuraterip(log)

    def deduct_and_score(self, log):
        """Process the accumulated deductions and score the log file."""
        # Deduct for all the per-track accumulated deductions.
        for error in log.track_errors:
            if log.track_errors[error]:
                log.add_deduction(error, len(log.track_errors[error]))

        super().deduct_and_score(log)



"""This module contains the functions which deal with drives and offsets."""

import re
import os
import sqlite3

from heybrochecklog.shared import get_path
from heybrochecklog import UnrecognizedException


def eval_offset(log, offset):
    """Validate the offset used by the ripped drive."""
    if not re.match('-?[0-9]+', offset):
        raise UnrecognizedException('Could not parse drive offset.')

    if not offset.startswith('-') and offset != '0':
        offset = '+' + offset

    if check_for_virtual_drives(log):
        log.add_deduction('Virtual drive')
        log.flagged = True
        return

    drivestr = prep_drive_name(log)
    if not drivestr:
        return

    results = drive_db_query(drivestr)
    if not results:
        # Drive not in database
        if offset == '0':
            log.add_deduction('Zero offset')
        log.unindexed_drive = True
        return

    offsets = {row[0] for row in results}
    if offset not in offsets:
        log.add_deduction(
            'Drive offset',
            extra_phrase='correct offsets are: {}'.format(', '.join(offsets)))


def check_for_virtual_drives(log):
    """Check for usage of virtual drives; they aren't good and should be reported."""
    fake_drives = [
        'Generic DVD-ROM SCSI CdRom Device'
        # TODO: Compile a more comprehensive list.
    ]
    if log.drive in fake_drives:
        return True
    return False


def prep_drive_name(log):
    """Prepare the drive name for a DB query."""
    drive = sub_drive_names(log.drive)
    drive_words = re.split(r'[^A-Za-z0-9]+', drive)
    drivestr = '%" AND Name LIKE "%'.join(drive_words)

    return drivestr


def sub_drive_names(drive):
    """Perform regex substitution actions on the drive name for better query results."""
    # Replace generic companies with real companies?
    drive = re.sub(r'JLMS', 'Lite-ON', drive)
    drive = re.sub(r'HL-DT-ST', 'LG Electronics', drive)
    drive = re.sub(r'Matshita', 'MATSHITA', drive)
    drive = re.sub(r'TSSTcorp(BD|CD|DVD)', r'TSSTcorp \1', drive)
    drive = re.sub(r'(\s+-\s|\s+)', ' ', drive)
    return drive


def drive_db_query(drivestr):
    """Query the SQLite3 DB for the drive offset."""
    db_path = os.path.join(get_path(), 'resources', 'drives.db')
    conn = sqlite3.connect(db_path)

    cursor = conn.cursor()
    cursor.execute('SELECT Offset FROM Drives WHERE Name LIKE "%' + drivestr + '%"')
    results = cursor.fetchall()

    conn.close()

    return results



"""This module contains the functions which deal with combined logs."""

import re
from heybrochecklog.logfile import LogFile


def split_combined(log):
    """Split a combined log into an array of logs. If there is a single
    log, return a list with one log inside it. Not relevant for XLD.
    """
    logs = []

    # Create a list of indices for combined log markers. By default # includes indices 0 and len()
    log_indices = [0] \
        + [i + 2 for i, line in enumerate(log.full_contents) if re.match(r'-{60}', line)] \
        + [len(log.full_contents)]

    # Split the log files. Create new log object for each log.
    for i, line in enumerate(log_indices):
        new_log = LogFile(log.full_contents[line:(log_indices[i + 1])], ripper=log.ripper)
        logs.append(new_log)

        # Return the array of logs if the end index of section is
        # equivalent to the length of the original log.
        if log_indices[i + 1] == max(log_indices):
            break

    return logs


def defragment(logs, eac95=False):
    """Re-combine split combined logs."""
    if len(logs) == 1 and not logs[0].htoa:
        return logs[0]

    full_contents = []
    for log in logs:
        full_contents += log.full_contents
    logs[0].full_contents = full_contents

    # Make sure HTOA CRC's match if they exist.
    if not eac95:
        htoa_logs = [log for log in logs if log.htoa]
        if htoa_logs:
            analyze_htoa(logs, htoa_logs)

    # Iterate through each log and patch the original log's data where new data exists.
    # logs[0] is used as the base log and returned at the end.
    for log in logs[1:]:
        # Make sure the albums have the same name.
        if log.album != logs[0].album:
            return logs[0]

        sub_settings(logs, log)
        sub_double_copy(logs, log)
        if not log.range:
            sub_track_errors(logs, log)

    # Remove HTOA detection if it was extracted.
    if not eac95:
        patch_htoa(logs, htoa_logs)

    if len(logs) > 1:
        logs[0].add_deduction('Combined log')

    return logs[0]


def sub_settings(logs, log):
    """Check settings of newer rip; add new deductions and pop fixed deductions."""
    # Remove the deductions from the first log if they aren't present in the final log,
    # but verify the number of tracks is consistent between the two.
    for deduction in logs[0].deductions.copy():
        if not log.has_deduction(deduction) and len(log.tracks) == len(logs[0].tracks):
            logs[0].remove_deduction(deduction)
    # Add additional deductions from the newer log to the older log.
    for deduction in log.deductions.copy():
        if not logs[0].has_deduction(deduction):
            logs[0].add_deduction(deduction)


def sub_double_copy(logs, log):
    """Check two copy only rips and score as T&C Rip."""
    if logs[0].has_deduction('Test & Copy') and len(logs[0].tracks) == len(log.tracks):
        for new_track, original_track in zip(log.tracks.values(), logs[0].tracks.values()):
            if 'copy crc' in new_track and 'copy crc' in original_track:
                if new_track['copy crc'] != original_track['copy crc']:
                    break
        else:
            logs[0].remove_deduction('Test & Copy')


def sub_track_errors(logs, log):
    """Substitute newer ripped track data for older ripped track data."""
    for track in log.tracks:
        # Don't substitute track errors for aborted copies.
        if 'copy crc' in log.tracks[track]:
            logs[0].tracks[track] = log.tracks[track]
            replace_accumulated_errors(track, logs, log)
            replace_crc_mismatches(track, logs, log)


def replace_accumulated_errors(track, logs, log):
    """Replace accumulated track errors."""
    for error in logs[0].track_errors:
        if track in log.track_errors[error]:
            continue
        for i, original_track in enumerate(logs[0].track_errors[error]):
            if track == original_track:
                logs[0].track_errors[error].pop(i)


def replace_crc_mismatches(track, logs, log):
    """Replace CRC mismatch errors."""
    if track not in log.crc_mismatch:
        for i, original_track in enumerate(logs[0].crc_mismatch):
            if track == original_track:
                logs[0].crc_mismatch.pop(i)


def analyze_htoa(logs, htoa_logs):
    """Analyze potential HTOA rips to and add deductions for CRCs and T&C."""
    # Remove htoa logs from logs list
    logs = [log for log in logs if log not in htoa_logs]

    if len(htoa_logs) >= 2:
        matches = [log for log in htoa_logs[1:] if log.tracks[0]['copy crc'] ==
                   htoa_logs[0].tracks[0]['copy crc']]
        if not matches:
            htoa_logs[0].add_deduction('CRC mismatch on HTOA extraction')
    elif len(htoa_logs) == 1:
        htoa_logs[0].add_deduction('HTOA not ripped twice')

    htoa_logs[0].remove_deduction('Test & Copy')
    logs.append(htoa_logs[0])


def patch_htoa(logs, htoa_logs):
    """Adjust the HTOA deductions based on the defragmented log."""
    htoa_ripped = any(log for log in htoa_logs if log.htoa_ripped)

    if htoa_ripped and logs[0].has_deduction('HTOA detected'):
        logs[0].remove_deduction('HTOA detected not extracted')

    if logs[0].has_deductions('Improper HTOA extraction', 'HTOA extracted'):
        logs[0].remove_deduction('HTOA extracted')



"""This module contains validation functions for log checking."""

from heybrochecklog import UnrecognizedException


def analyze_accuraterip(log):
    """Analyze the AccurateRip results in the log."""
    if log.accuraterip:
        for ar_result in log.accuraterip:
            if (ar_result[0] != log.accuraterip[0][0]
                    and (ar_result[1] is not None and log.accuraterip[0][1] is not None)
                    and (int(ar_result[1]) >= 5 or int(log.accuraterip[0][1]) >= 5)):
                log.add_deduction('AccurateRip discrepancies')
                break
    elif any('copy crc' in data for tnum, data in log.tracks.items()):
        log.add_deduction('AccurateRip')


def check_crc_mismatch(log, track_num, track_data):
    """Check a track block for a CRC mismatch."""
    if all(data in track_data for data in ['test crc', 'copy crc']):
        if track_data['test crc'] != track_data['copy crc']:
            log.crc_mismatch.append(track_num)


def validate_track_count(log):
    """Verify the presence of all tracks and check for a data track."""
    if not log.range:
        # Data tracks have one extra ToC entry, but one less ripped track
        if max(log.tracks.keys()) + 1 == max(log.toc.keys()):
            log.add_deduction('Data track detected')
        elif len(log.tracks) != len(log.toc):
            if not log.ripper == 'XLD' or not log.has_deduction('HTOA extracted'):
                raise UnrecognizedException('Not all tracks are represented in the log')


def validate_track_settings(log, xld=False):
    """Also verify that each track contains the required data."""
    if xld:
        if log.range:
            required_settings = ['copy crc']
        else:
            required_settings = ['filename', 'copy crc']
    else:
        required_settings = ['filename', 'peak', 'copy crc']

    for track in log.tracks:
        if track in log.track_errors['Aborted copy']:
            pass
        elif not all(setting in log.tracks[track] for setting in required_settings):
            raise UnrecognizedException('Unable to confirm presence of required track data')

    # Check for Test & Copy and CRC Mismatches
    if not all('test crc' in track for track in log.tracks.values()):
        if log.has_deduction('HTOA extracted') or log.has_deduction('HTOA not ripped twice'):
            # Verify that every track minus HTOA is T&C (Range based is index 0)
            if all('test crc' in track for i, track in log.tracks.items() if i):
                return
        log.add_deduction('Test & Copy')



"""This module handles the log translation functionality of the heybrochecklog package."""

import re
from collections import OrderedDict

from heybrochecklog import UnrecognizedException
from heybrochecklog.analyze import analyze_log
from heybrochecklog.shared import get_log_contents, open_json
from heybrochecklog.logfile import LogFile


def translate_log(log_file):
    """Initialize and capture all logs."""
    try:
        contents = get_log_contents(log_file)
        log = LogFile(contents)
        return translate_wrapper(log)
    except UnicodeDecodeError:
        return {'unrecognized': 'Could not decode log'}


def translate_log_from_contents(contents):
    """Translate a log file given its contents."""
    log = LogFile(contents.split('\n'))
    try:
        return translate_wrapper(log)
    except UnicodeDecodeError:
        return {'unrecognized': 'Could not decode log'}


def translate_wrapper(log):
    """Translate log given log file."""
    try:
        analyze_log(log)
    except UnrecognizedException as exception:
        return {
            'unrecognized': str(exception),
            'log': ''.join(log.full_contents)
        }

    if log.language == 'english':
        return {
            'unrecognized': False,
            'language': 'english',
            'log': ''.join(log.full_contents)
        }

    return sub_english(log)


def sub_english(log):
    """Translate the log file and return a dict of info and log."""
    english = open_json('eac', 'english.json')['translation']
    foreign = open_json('eac', '{}.json'.format(log.language))['translation']

    # Sort foreign lines from longest to shortest
    foreign = OrderedDict(sorted(foreign.items(), key=lambda t: len(t[1][0]), reverse=True))

    # Compile all the regex now instead of repeating.
    for key, value in foreign.items():
        for i, v in enumerate(value):
            value[i] = re.escape(v)
        foreign[key] = re.compile('|'.join(value), flags=re.IGNORECASE)

    # Iterate through each line and find/replace each string.
    new_log = []
    for line in log.full_contents:
        if not line:  # No use wasting time here.
            new_log.append('')
        else:
            for key, regex in foreign.items():
                if regex.search(line):
                    for value in english[key]:
                        line = regex.sub(value, line)
            new_log.append(line)

    re_space_settings(new_log)
    new_log = ''.join(new_log)

    return {
        'unrecognized': False,
        'language': log.language,
        'log': new_log
    }


def re_space_settings(log):
    """Fix the spacing in the rip settings block."""
    spacings = [24, 44, 32]  # Number of characters before colons in each block of settings.
    end_of_blocks = ['Make use of C2 pointers', 'Gap handling']  # Last setting in blocks.
    index = 0

    for i, line in enumerate(log):
        if re.match('Used drive', line):
            continue  # Ignore that line

        result = re.search(r'^([A-Z][A-Za-z0-9- ]+[a-z]) +:(.*)$', line)
        if result:
            word = result.group(1)
            spaces = spacings[index] - len(word)
            log[i] = word + ' ' * spaces + ':' + result.group(2) + '\n'

            if word in end_of_blocks:
                index += 1

        if re.match('TOC', line):  # No need to continue after settings are done.
            return



"""This module contains the generation functions for dicts of to-match lines
when marking up log files.
"""

import re
import html
from heybrochecklog.shared import format_pattern as fmt_ptn


def eac_track_matches(translation):
    """Generate the list of to-match lines from translations."""
    source_one = {
        'log4': ['1269', '1270', '1217', '1299', '1227', '1218'],
        'good': ['1220', '1281'],
        'crc': ['1271', '1272']
    }
    source_two = {
        'log3': ['1221'],
        'badish': ['1330', '1283', '1210', '1211'],
        'bad': ['1212', '1213', '1214', '1215', '1216', '1228']
    }
    return {
        **generate_match_type(translation, source_one),
        **generate_match_type(translation, source_two, append='.*')
    }


def xld_track_matches():
    """Return a dictionary containing the XLD matches."""
    return {
        'full_line': [
            ['log4', 'Statistics'],
            ['good', '-&gt;Accurately ripped'],
            ['badish', '-&gt;Track not present in AccurateRip database'],
            ['bad', '-&gt;Rip may not be accurate'],
            ['bad', 'List of damaged sector positions +:'],
            ['badish', r'\(\d+\) \d{2}:\d{2}:\d{2}'],
            ['log3', r'\/.+\.(?:[Ff][Ll][Aa][Cc]|[Ww][Aa][Vv]|[Mm][Pp]3|[Aa][Aa][Cc])']
        ],
        'crc': [
            r'CRC32 hash \(test run\)',
            'CRC32 hash'
        ],
        'statistics': {
            'bad': [
                'Read error',
                r'Skipped \(treated as error\)',
                'Inconsistency in error sectors',
                'Damaged sector count'
            ],
            'badish': [
                r'Jitter error \(maybe fixed\)',
                'Retry sector count',
                r'Edge jitter error \(maybe fixed\)',
                r'Atom jitter error \(maybe fixed\)',
                r'Drift error \(maybe fixed\)',
                r'Dropped bytes error \(maybe fixed\)',
                r'Duplicated bytes error \(maybe fixed\)'
            ]
        }
    }


def eac_footer_matches(translation):
    """Matches for the EAC footer block."""
    source = {
        'good': ['1336', '1222', '1225'],
        'badish': ['1333', '1334', '1344', '1335'],
        'bad': ['1284', '1224'],
        'log4 log5': ['1275']
    }
    matches = generate_match_type(translation, source)

    # AccurateRip stuff
    source = {
        'good': ['1340'],
        'badish': ['1339', '1341'],
        'bad': ['1342', '1343']
    }
    matches = generate_match_type(translation, source, matches=matches, prepend='.+')

    if '1290' in translation:
        ar_prepend = r'{} +\d+ +'.format(translation['1290'])
        source = {
            'good': ['1277'],
            'badish': ['1279'],
            'bad': ['1276', '1278']
        }
        matches = generate_match_type(translation, source, matches=matches, prepend=ar_prepend,
                                      append='.+')

    # Checksum stuff
    if '1325' in translation:
        matches['good'].append('==== {} [0-9A-F]+ ===='.format(translation['1325']))

    # EAC HAS A TYPO FOR "NO ERRORS OCCURED" WTF
    if translation['1222'] == 'No errors occurred':
        matches['good'].append('No errors occured')

    return matches


def xld_footer_matches():
    """Matches for the XLD footer block."""
    return {
        'good': [
            'No errors occurred',
            'End of status report'
        ],
        'badish': [
            'Some inconsistencies found'
        ]
    }


def xld_ar_summary():
    """Matches for the XLD AccurateRip Summary block."""
    return {
        'good': [
            r'Track \d+ : OK.+',
            html.escape('-&gt;All tracks accurately ripped.*')
        ],
        'badish': [
            r'Track \d+ : NG.+',
            'Disc not found in AccurateRip DB',
            r'-&gt;\d+ tracks? accurately ripped, \d+ tracks? not'
        ],
        'log4 log5': ['AccurateRip Summary']
    }


def generate_match_type(translation, source, matches=None, prepend='', append=''):
    """Function to generate the match types."""
    matches = {} if not matches else matches
    for match_type in source.keys():
        if match_type not in matches:
            matches[match_type] = []
        for line_id in source[match_type]:
            if line_id in translation:
                match = prepend + re_paren(translation[line_id]) + append
                matches[match_type].append(html.escape(match))

    return matches


def re_paren(line):
    """Regex the comma. Quality docstring."""
    line = re.sub(r'\(', r'\(', fmt_ptn(line))
    return re.sub(r'\)', r'\)', line)



"""This module contains functions which parse lines for data."""

import re
from heybrochecklog.resources import VERSIONS
from heybrochecklog import UnrecognizedException
from heybrochecklog.shared import format_pattern as fmt_ptn


def index_toc(log):
    """Index the ToC data of the log."""
    re_toc = re.compile(r' ([0-9]+) \| [0-9:\.]+ \| [0-9:\.]+ \| ([0-9]+) \| ([0-9]+)')
    for line in log.contents[log.index_toc:log.index_tracks]:
        result = re_toc.search(line)
        if result:
            log.toc[int(result.group(1))] = [int(result.group(2)), int(result.group(3))]


def get_track_number(log, index, track_word):
    """Get the track number from the header line of a track block."""
    result = re.search(r'{} ([0-9]+)'.format(fmt_ptn(track_word)), log.contents[index])
    if result:
        return int(result.group(1))
    elif log.range:  # EAC range rip has no track number
        return 0
    else:
        raise UnrecognizedException('A track has an invalid block header')


def parse_settings(track_data, track_settings, line):
    """Loop through and parse the settings used in the rip."""
    for setting, reg in track_settings.items():
        result = reg.match(line)
        if result:
            track_data[setting] = result.group(1)


def parse_accuraterip(log, ar_patterns, line):
    """Parse line for an AccurateRip result."""
    for status, re_accurip in ar_patterns:
        result = re.search(fmt_ptn(re_accurip), line)
        if result and result.lastindex == 1:
            log.accuraterip.append([status, result.group(1)])
        elif result and result.lastindex is None:
            log.accuraterip.append([status, None])


def parse_range_accuraterip(log, ar_rr_patterns):
    """Parse range rip footer for AccurateRip results."""
    for line in log.contents[log.index_footer:]:
        parse_accuraterip(log, ar_rr_patterns, line)


def parse_errors_eac(log, err_patterns, track_num, line):
    """Parse line of an EAC log for a ripping error."""
    for error, re_err in err_patterns:
        if (track_num not in log.track_errors[error]
                and re.match(r' ' + fmt_ptn(re_err), line)):
            log.track_errors[error].append(track_num)


def parse_errors_xld(log, err_patterns, track_num, line):
    """Parse line of a XLD log for a ripping error."""
    for error, re_err in err_patterns:
        if track_num not in log.track_errors[error]:
            result = re.search(r' ' + fmt_ptn(re_err) + r' : ([0-9]+)', line)
            if result and result.group(1) != "0":
                log.track_errors[error].append([track_num, int(result.group(1))])


def parse_checksum(log, regex, imp_version, deduc_line):
    """Parse line(s) for presence of a checksum."""
    re_checksum = re.compile(fmt_ptn(regex))
    for line in log.contents[log.index_footer:]:
        if re_checksum.match(line):
            log.checksum = True
            break
    else:  # If checksum not found
        # Compare version numbers to see if Log is older than checksums.
        for version in VERSIONS[log.ripper]:
            if version[0] == log.version:
                log_version = version
            if version[0] == imp_version:
                imp_version = version
        if VERSIONS[log.ripper].index(log_version) <= VERSIONS[log.ripper].index(imp_version):
            log.add_deduction('Checksum')
        else:
            log.add_deduction(deduc_line + ' (no checksum)')



"""This module contains the EAC Version <=0.95 Log Checker."""

import re
from heybrochecklog import UnrecognizedException
from heybrochecklog.shared import format_pattern as fmt_ptn
from heybrochecklog.markup import markup
from heybrochecklog.score.logchecker import LogChecker
from heybrochecklog.score.modules import combined, drives, parsers, validation


class EAC95Checker(LogChecker):
    """This class analyzes <=0.95 EAC Log Files."""

    def check(self, main_log):
        """Checks the EAC logs."""
        logs = combined.split_combined(main_log)
        for log in logs:
            if len(log.concat_contents) < 12:
                raise UnrecognizedException('Cannot parse log file; log file too short')

            log.version = 'EAC <=0.95'
            log.album = log.concat_contents[1]
            log.drive = self.check_drive(log)

            self.index_log(log, ninety_five=True)
            self.evaluate_settings(log)
            self.check_tracks(log)
            if self.markup:
                markup(log, self.patterns, self.translation)

        main_log = combined.defragment(logs, eac95=True)
        validation.validate_track_settings(main_log)
        self.deduct_and_score(main_log)

        return main_log

    def check_drive(self, log):
        """Check the drive of the log and verify it is an allowed drive."""
        regex = r' ?: (.*) Adapter:[ 0-9]+ID:[ 0-9]+$'
        return self.get_drive(regex, log.concat_contents[2])

    def all_range_index(self, log, line):
        """Match the Range Rip line in the log file."""
        if log.index_tracks is None and re.match(fmt_ptn(self.patterns['range']), line):
            return True
        return False

    def all_range_index_action(self, log, line_num):
        """Action to take when the range rip line is matched."""
        log.track_indices.append(line_num)
        log.range = True

    def evaluate_settings(self, log):
        """Evaluate the log for usage of proper rip settings.
        Overwriting the base class for different 0.95 behavior.
        """
        psettings = self.patterns['settings']
        full_psettings = self.patterns['full line settings']
        proper_settings = self.patterns['proper settings']

        # Compile regex beforehand
        settings, full_settings = {}, {}
        for key, regex in psettings.items():
            settings[key] = re.compile(fmt_ptn(regex))
        for key, regex in full_psettings.items():
            full_settings[key] = re.compile(fmt_ptn(regex) + ' : (.*)')

        # Iterate through line in the settings, and verify each setting in `settings` dict
        for line in log.contents[log.index_settings:log.index_tracks]:
            for key, setting in list(settings.items()):
                result = setting.search(line)
                if result:
                    if key == 'Drive offset':
                        offset = re.search(r'.+: ([-0-9]+)', line)
                        drives.eval_offset(log, offset.group(1))
                    del settings[key]
            for key, setting in list(full_settings.items()):
                result = setting.search(line)
                if result:
                    if not re.search(fmt_ptn(proper_settings[key]), result.group(1)):
                        log.add_deduction(key)
                    del full_settings[key]
                    break

            self.check_bad_settings(log, line)

        self.evaluate_unmatched_settings(log, settings)

    def check_offset(self, log, line, off_settings):
        """Check a log file line for proper offset."""
        found = False
        for key, setting in off_settings.items():
            result = setting.search(line)
            if result:
                if key == 'Combined offset':
                    log.add_deduction('Combined offset')
                else:
                    drives.eval_offset(log, result.group(1))
                found = True

        return found

    def check_bad_settings(self, log, line):
        """Evaluate the instant -100 point deductions
        (destructive normalization and compression offset)."""
        bad_settings = self.patterns['bad settings']
        for sett, pattern in bad_settings.items():
            if re.search(fmt_ptn(pattern), line):
                log.add_deduction(sett)

    def evaluate_unmatched_settings(self, log, settings):
        """Evaluate all unmatched settings and deduct for them.
        <=0.95 is using a match/no match string algorithm, so it's a deduction if no match."""
        if log.has_deduction('Combined offset') and 'Drive offset' in settings:
            del settings['Drive offset']
        for key in settings:
            log.add_deduction(key)

    def check_tracks(self, log):
        """Get track data for each track and check for errors."""
        tsettings = self.patterns['track settings']
        track_settings = {
            'filename': re.compile(r' ' + fmt_ptn(tsettings['filename']) + r' (.*)'),
            'pregap': re.compile(r' ' + fmt_ptn(tsettings['pregap']) + r' ([0-9:\.]+)'),
            'peak': re.compile(r' ' + fmt_ptn(tsettings['peak']) + r' ([0-9\.])+ %'),
            'test crc': re.compile(r' ' + fmt_ptn(tsettings['test crc']) + r' ([A-Z0-9]{8})'),
            'copy crc': re.compile(r' ' + fmt_ptn(tsettings['copy crc']) + r' ([A-Z0-9]{8})')
        }

        self.analyze_tracks(log, track_settings, parsers.parse_errors_eac, accuraterip=False)

    def evaluate_tracks(self, log):
        """Evaluate the analyzed track data for deficiencies."""
        # Deduct for a Range Rip
        if log.range:
            log.add_deduction('Range rip')

    def deduct_and_score(self, log):
        """Process the accumulated deductions and score the log file."""
        # EAC <=0.95 mandatory deductions.
        log.add_deduction('EAC 0.95')
        log.add_deduction('EAC <1.0 (no checksum)')
        if not log.range:
            log.add_deduction('Gap handling')
        else:
            log.add_deduction('Could not verify gap handling')

        # Deduct for all the per-track accumulated deductions.
        for error in log.track_errors:
            if log.track_errors[error]:
                log.add_deduction(error, len(log.track_errors[error]))

        super().deduct_and_score(log)



class UnrecognizedException(Exception):
    pass


import argparse  # noqa: E402
from pathlib import Path  # noqa: E402
from heybrochecklog.score import score_log  # noqa: E402
from heybrochecklog.translate import translate_log  # noqa: E402


def parse_args():
    """Parse arguments."""
    description = 'Tool to analyze, translate, and score a CD Rip Log.'

    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('log', help='log file to check.', nargs='+')
    parser.add_argument('-t', '--translate', help='translate a foreign log to English', action='store_true')
    parser.add_argument('-m', '--markup', help='print the marked up version of the log after analyzing',
                        action='store_true')
    parser.add_argument('-s', '--score-only', help='Only print the score of the log.', action='store_true')

    return parser.parse_args()


def runner():
    """Main function to handle command line usage of the heybrochecklog package."""
    args = parse_args()
    for log_path in args.log:
        log_file = Path(log_path)
        if not log_file.is_file():
            print('{} does not exist.'.format(log_path))
        elif args.translate:
            translate_(args, log_file, log_path)
        elif args.log:
            score_(args, log_file, log_path)


def score_(args, log_file, log_path):
    log = score_log(log_file, args.markup)
    if args.score_only:
        if not log['unrecognized']:
            print(log['score'])
        else:
            print('Log is unrecognized: {}'.format(log['unrecognized']))
    else:
        try:
            print(format_score(log_path, log, args.markup))
        except UnicodeEncodeError as error:
            print('Cannot encode logpath: {}'.format(error))


def translate_(args, log_file, log_path):
    log = translate_log(log_file)
    try:
        print(format_translation(log_path, log))
    except UnicodeEncodeError as error:
        print('Cannot encode logpath: {}'.format(error))


def format_score(logpath, log, markup):
    """Turn a log file JSON into a pretty string."""
    output = []
    output.append('\nLog: ' + logpath)
    if log['unrecognized']:
        output.append('\nLog is unrecognized: {}'.format(log['unrecognized']))
    else:
        if log['flagged']:
            output.append('\nLog is flagged: {}'.format(log['flagged']))
        output.append('\nDisc name: {}'.format(log['name']))
        output.append('\nScore: {}'.format(log['score']))

        if log['deductions']:
            output.append('\nDeductions:')
            for deduction in log['deductions']:
                output.append('  >>  {}'.format(deduction[0]))

        if markup:
            output.append('\n\nStylized Log:\n\n{}'.format(log['contents']))

    return '\n'.join(output)


def format_translation(logpath, log):
    """Turn a translated log JSON into a pretty string."""
    output = []
    output.append('\nLog: ' + logpath)

    if log['unrecognized']:
        output.append('\nFailed to recognize log. {}'.format(log['unrecognized']))
    elif log['language'] == 'english':
        output.append('\nLog is already in English!')
    else:
        output.append('\nOriginal language: {}'.format(log['language']).title())
        output.append('\n---------------------------------------------------')
        output.append('\n' + log['log'])

    return '\n'.join(output)



#!/usr/bin/env python3
import sys
from setuptools import setup
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


with open('README.md') as desc:
    long_description = desc.read()


setup(
    name='heybrochecklog',
    version='1.2',
    description='A python tool for evaluating and working with EAC/XLD CD rip logs.',
    long_description=long_description,
    long_description_content_type='text/markdown/',
    license='GPLv3',
    author='lights',
    author_email='lights@tutanota.de',
    url='https://github.com/ligh7s/hey-bro-check-log',
    keywords='logchecker eac xld cd rips',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
    ],
    packages=[
        'heybrochecklog',
        'heybrochecklog.resources',
        'heybrochecklog.resources.eac',
        'heybrochecklog.resources.eac95',
        'heybrochecklog.markup',
        'heybrochecklog.score',
        'heybrochecklog.score.modules',
    ],
    include_package_data=True,
    python_requires='>=3.5',
    tests_require=['pytest'],
    cmdclass={'test': PyTest},
    install_requires=[
        'chardet'
    ],
    entry_points={
        'console_scripts': [
            'heybrochecklog=heybrochecklog:runner',
        ],
    },
)



"""A module which contains the base log checker, with the code shared between
more specific log checkers.
"""

import re

from heybrochecklog.resources import VERSIONS
from heybrochecklog import UnrecognizedException
from heybrochecklog.shared import format_pattern as fmt_ptn
from heybrochecklog.score.modules import parsers, validation, drives


class LogChecker:
    """The base log checker to be subclassed by more specific log checkers."""

    def __init__(self, patterns, translation=None, markup=False):
        self.patterns = patterns
        self.translation = translation
        self.markup = markup

    def verify_version(self, regex, line, ripper):
        """Verify that the version of the log is legitimate."""
        result = regex.search(line)
        if result:
            version, date = result.group(1), result.group(2)
            if (version, date) in VERSIONS[ripper]:
                return version
        raise UnrecognizedException('Unrecognized {} version'.format(ripper))

    def get_drive(self, regex, line):
        """Get the name of the ripping drive used."""
        re_drive = re.compile(fmt_ptn(self.patterns['drive']) + regex)
        result = re_drive.match(line)
        if result:
            return result.group(1).strip()
        raise UnrecognizedException('Could not parse ripping drive')

    def index_log(self, log, ninety_five=False):
        """Index key line numbers inside the log."""
        if ninety_five:
            read_mode = re.compile(re.sub(' +', ' ', fmt_ptn(self.translation['1234'])))
        else:
            read_mode = re.compile(fmt_ptn(self.patterns['settings']['Read mode']))
        toc = re.compile(fmt_ptn(self.patterns['toc'])) if 'toc' in self.patterns else None

        for i, line in enumerate(log.contents):
            if log.index_settings is None and read_mode.match(line):
                log.index_settings = i
            elif log.index_toc is None and 'toc' in self.patterns and toc.match(line):
                log.index_toc = i
            elif self.all_range_index(log, line):
                self.all_range_index_action(log, i)
            elif re.match(fmt_ptn(self.patterns['track']) + r' [0-9]+$', line):
                log.track_indices.append(i)

        self.validate_indices(log)

    def validate_indices(self, log):
        """Validate the indices of notable lines in the log."""
        if not log.track_indices:
            raise UnrecognizedException('No tracks found')

        # +4 to compensate for range rip lines
        for i in range(max(log.track_indices) + 4, len(log.contents)):
            # Need to check the full_contents and add the extra space for AR summary leading space.
            if re.match(r' ?\w', log.full_contents[i]):
                log.track_indices.append(i - 1)
                log.index_footer = i
                break

        log.index_tracks = min(log.track_indices)
        if not log.index_toc:
            log.index_toc = log.index_tracks

    def all_range_index(self, log, line):
        """Match the All Tracks or Range Rip line, depending on subclassed ripper."""
        pass

    def all_range_index_action(self, log, line_num):
        """Action to take when detection of the All Tracks or Range Rip line occurs."""
        pass

    def evaluate_settings(self, log):
        """Evaluate the log for usage of proper rip settings."""
        psettings = self.patterns['settings']
        proper_settings = self.patterns['proper settings']

        # Compile regex beforehand
        settings = {}
        for key, setting in psettings.items():
            settings[key] = re.compile(fmt_ptn(setting) + r' : (.*)')

        # Iterate through line in the settings, and verify each setting in `sets` dict
        for line in log.contents[log.index_settings:log.index_toc]:
            for key, setting in list(settings.items()):
                result = setting.search(line)
                if result and key == 'Drive offset':
                    drives.eval_offset(log, result.group(1))
                    del settings[key]
                elif result:
                    if not re.search(fmt_ptn(proper_settings[key]), result.group(1)):
                        log.add_deduction(key)
                    del settings[key]
                    break

            self.check_bad_settings(log, line)

        self.evaluate_unmatched_settings(log, settings)

    def check_bad_settings(self, log, line):
        """Evaluate the bad settings, override in subclass if desired."""
        pass

    def evaluate_unmatched_settings(self, log, settings):
        """Evaluate all unmatched settings and deduct for them."""
        for key in settings:
            if log.range and key == 'Gap handling':
                log.add_deduction('Could not verify gap handling')
            elif key == 'Gap handling':
                log.add_deduction('Gap handling')
            elif key == 'ID3 tags':
                log.add_deduction('Could not verify presence of ID3 tags')
            elif key == 'Null samples':
                log.add_deduction('Could not verify usage of null samples')
            else:
                raise UnrecognizedException('One or more required settings could not be found')

    def analyze_tracks(self, log, track_settings, parse_errors, accuraterip=True):
        """Get track data for each track and check for errors."""
        ar_patterns = self.patterns['accuraterip'].items() if accuraterip else {}
        err_patterns = self.patterns['track errors'].items()

        for i, index in enumerate(log.track_indices):
            track_data = {}
            track_num = parsers.get_track_number(log, index, self.patterns['track'])

            for line in log.contents[log.track_indices[i]:log.track_indices[i + 1]]:
                # Collect the track data using the track_settings list.
                parsers.parse_settings(track_data, track_settings, line)
                # Log the AccurateRip results - add AR status to log.accuraterip list.
                if accuraterip:
                    parsers.parse_accuraterip(log, ar_patterns, line)
                # Ripping Errors - Loop through errors in json track errors.
                parse_errors(log, err_patterns, track_num, line)

            validation.check_crc_mismatch(log, track_num, track_data)

            log.tracks[track_num] = track_data
            if log.track_indices[i + 1] == max(log.track_indices):
                break

        self.evaluate_tracks(log)

    def evaluate_tracks(self, log):
        """Evaluate the analyzed track data for deficiencies (actually split off the
        logchecker-specific) stuff ;)
        """
        pass

    def deduct_and_score(self, log):
        """Process the accumulated deductions and score the log file."""
        if log.crc_mismatch:
            log.add_deduction('CRC mismatch', len(log.crc_mismatch))

        # Sum all the deductions and calculate score
        log.score -= sum([de[1] for de in log.deductions.values() if isinstance(de[1], int)])
